# Copyright 2022 Scott K Logan
# Licensed under the Apache License, Version 2.0

import asyncio
from asyncio.events import AbstractEventLoop
from asyncio.events import get_event_loop
from asyncio.tasks import Task
from asyncio.transports import WriteTransport
from contextlib import AbstractAsyncContextManager
from typing import Any
from typing import Dict
from typing import Optional
from typing import Tuple
import xml.etree.ElementTree as Et

from aioraven.data import RAVEnData
from aioraven.device import RAVEnBaseDevice
from aioraven.device import RAVEnConnectionError
from aioraven.device import RAVEnNotOpenError
from aioraven.protocols import RAVEnReaderProtocol
from aioraven.reader import RAVEnReader


class RAVEnWriter:
    """Write commands to a RAVEn device."""

    def __init__(
        self,
        transport: WriteTransport,
        protocol: RAVEnReaderProtocol
    ) -> None:
        """
        Construct a RAVEnWriter.

        :param transport: The transport instance to wrap.
        :param protocol: The reader protocol for the connection.
        """
        self._transport = transport
        self._protocol = protocol

    def __repr__(self) -> str:
        info = [self.__class__.__name__]
        info.append('transport=%r' % self._transport)
        return '<%s>' % ' '.join(info)

    def write_cmd(
        self,
        cmd_name: str,
        args: Optional[Dict[str, str]] = None,
    ) -> None:
        element_cmd = Et.Element('Command')
        element_name = Et.SubElement(element_cmd, 'Name')
        element_name.text = cmd_name
        for k, v in (args or {}).items():
            element_arg = Et.SubElement(element_cmd, k)
            element_arg.text = v
        tree = Et.ElementTree(element_cmd)
        tree.write(self._transport, encoding='ASCII', xml_declaration=False)

    def abort(self) -> None:
        self._transport.abort()

    def close(self) -> None:
        self._transport.close()

    async def wait_closed(self) -> None:
        await self._protocol._get_close_waiter(self)


async def open_connection(
    host: str,
    port: int,
    *,
    loop: Optional[AbstractEventLoop] = None,
) -> Tuple[RAVEnReader, RAVEnWriter]:
    """
    Establish a network connection to a RAVEn device.

    Additional optional keyword arguments are passed to
    `AbstractEventLoop.create_connection()`.

    :param host: The hostname or IP address to connect to.
    :param port: The TCP port number to connect to.
    :param loop: The event loop instance to use.
    """
    if loop is None:
        loop = get_event_loop()
    reader = RAVEnReader(loop=loop)
    protocol = RAVEnReaderProtocol(reader, loop=loop)
    try:
        transport, _ = await loop.create_connection(
            lambda: protocol, host=host, port=port)
    except OSError as ex:
        raise RAVEnConnectionError(f'{ex}') from ex
    writer = RAVEnWriter(transport, protocol)
    return reader, writer


class RAVEnStreamDevice(
    RAVEnBaseDevice,
    AbstractAsyncContextManager['RAVEnStreamDevice']
):
    """Read and write coordination for stream-based RAVEn devices."""

    _reader: Optional[RAVEnReader] = None
    _writer: Optional[RAVEnWriter] = None

    async def open(self) -> None:
        raise NotImplementedError()

    async def abort(self) -> None:
        if self._writer:
            self._writer.abort()
            try:
                await self._writer.wait_closed()
            except IOError:
                pass
            self._reader = None
            self._writer = None

    async def close(self) -> None:
        if self._writer:
            self._writer.close()
            try:
                await self._writer.wait_closed()
            except IOError:
                pass
            self._reader = None
            self._writer = None

    async def synchronize(
        self, *, retries: int = 2, timeout: float = 1.0,
    ) -> None:
        await asyncio.sleep(0.05)
        for _try in range(retries, -1, -1):
            try:
                # Try a few times to communicate with the device,
                # allowing any data already in the buffer to flush.
                await asyncio.wait_for(self.get_meter_list(), timeout)
            except (RAVEnConnectionError, asyncio.TimeoutError):
                if not _try:
                    raise
            else:
                break

    async def _query(
        self,
        cmd_name: str,
        res_name: Optional[str] = None,
        args: Optional[Dict[str, str]] = None,
    ) -> Optional[RAVEnData]:
        if not self._reader or not self._writer:
            raise RAVEnNotOpenError()
        waiter: Optional[Task[Optional[RAVEnData]]] = None
        if res_name:
            waiter = asyncio.create_task(self._reader.read_tag(res_name))
            await asyncio.sleep(0)
        try:
            self._writer.write_cmd(cmd_name, args)
            return await waiter if waiter is not None else None
        except (Et.ParseError, IOError) as ex:
            raise RAVEnConnectionError(f'{ex}') from ex

    async def __aenter__(self) -> 'RAVEnStreamDevice':
        await self.open()
        return self

    async def __aexit__(
        self,
        exc_type: Any,
        exc_value: Any,
        traceback: Any
    ) -> None:
        await self.close()


class RAVEnNetworkDevice(RAVEnStreamDevice):
    """A network-connected RAVEn device."""

    def __init__(
        self,
        host: str,
        port: int,
        *,
        loop: Optional[AbstractEventLoop] = None,
    ) -> None:
        """
        Construct a RAVEnNetworkDevice.

        Additional optional keyword arguments are passed to
        `AbstractEventLoop.create_connection()`.

        :param host: The hostname or IP address to connect to.
        :param port: The TCP port number to connect to.
        :param loop: The event loop instance to use.
        """
        self._host = host
        self._port = port
        self._loop = loop

    def __repr__(self) -> str:
        info = [self.__class__.__name__]
        info.append('host=%s' % self._host)
        info.append('port=%s' % self._port)
        return '<%s>' % ' '.join(info)

    async def open(self) -> None:
        """Open the connection to the RAVEn device."""
        if self._reader or self._writer:
            return
        self._reader, self._writer = await open_connection(
            host=self._host, port=self._port, loop=self._loop)



# Copyright 2022 Scott K Logan
# Licensed under the Apache License, Version 2.0

import asyncio
import os

from aioraven.data import MeterList
from aioraven.device import RAVEnConnectionError
from aioraven.device import RAVEnNotOpenError
from aioraven.serial import RAVEnSerialDevice
import pytest

from .mock_device import mock_pty_device


if os.name == 'nt':
    pytest.skip(
        'Pseudo-terminals are not supported on non-Unix platforms',
        allow_module_level=True)


@pytest.mark.asyncio
async def test_serial_data():
    """Verify simple device query behavior."""
    responses = {
        b'<Command><Name>get_meter_list</Name></Command>':
            b'<MeterList>'
            b'    <DeviceMacId>0x0123456789abcdef</DeviceMacId>'
            b'</MeterList>',
    }

    async with mock_pty_device(responses) as mock_device:
        async with RAVEnSerialDevice(mock_device) as dut:
            actual = await dut.get_meter_list()

    assert actual == MeterList(
        device_mac_id=bytes.fromhex('0123456789ABCDEF'),
        meter_mac_ids=[])


@pytest.mark.asyncio
async def test_serial_disconnect():
    """Verify behavior when a device is unexpectedly disconnected."""
    responses = {
        b'<Command><Name>get_meter_list</Name></Command>':
            b'<MeterList>'
            b'    <DeviceMacId>0x0123456789abcdef</DeviceMacId>'
            b'</MeterList>',
    }

    async with mock_pty_device(responses) as mock_device:
        dut = RAVEnSerialDevice(mock_device)
        await dut.open()
        assert await dut.get_meter_list()
    async with dut:
        with pytest.raises(RAVEnConnectionError):
            await dut.get_meter_list()
        with pytest.raises(RAVEnConnectionError):
            await dut.get_meter_list()


@pytest.mark.asyncio
async def test_serial_incomplete():
    """Verify behavior when a partial fragment is received."""
    responses = {
        b'<Command><Name>get_meter_list</Name></Command>':
            b'<MeterList>'
            b'    <DeviceMacId>0x0123456789abcdef</DeviceMacId>'
            b'</MeterList>'
            b'<DeviceInfo>'
    }

    async with mock_pty_device(responses) as mock_device:
        dut = RAVEnSerialDevice(mock_device)
        await dut.open()
        assert await dut.get_meter_list()

        task = asyncio.create_task(dut.get_meter_list())
        await asyncio.wait((task,), timeout=0.05)
    async with dut:
        with pytest.raises(RAVEnConnectionError):
            assert not await task


@pytest.mark.asyncio
async def test_serial_not_open():
    """Verify behavior when reading from an unopened device."""
    async with mock_pty_device({}) as mock_device:
        dut = RAVEnSerialDevice(mock_device)
        with pytest.raises(RAVEnNotOpenError):
            await dut.get_device_info()


@pytest.mark.asyncio
async def test_serial_parse_error():
    """Verify behavior when invalid syntax is received."""
    responses = {
        b'<Command><Name>get_device_info</Name></Command>':
            b'</DeviceInfo>',
        b'<Command><Name>get_meter_list</Name></Command>':
            b'<MeterList>'
            b'    <DeviceMacId>0x0123456789abcdef</DeviceMacId>'
            b'</MeterList>',
    }

    async with mock_pty_device(responses) as mock_device:
        async with RAVEnSerialDevice(mock_device) as dut:
            with pytest.raises(RAVEnConnectionError):
                await dut.get_device_info()
            actual = await dut.get_meter_list()

    assert actual == MeterList(
        device_mac_id=bytes.fromhex('0123456789ABCDEF'),
        meter_mac_ids=[])


@pytest.mark.asyncio
async def test_serial_repr():
    """Verify representation of a serial device."""
    async with mock_pty_device({}) as mock_device:
        async with RAVEnSerialDevice(mock_device) as dut:
            assert 'RAVEnSerialDevice' in str(dut)


@pytest.mark.asyncio
async def test_serial_timeout_recovery():
    """Verify device recovery after a timeout."""
    responses = {
        b'<Command><Name>get_device_info</Name></Command>':
            b'<DeviceInfo>'
            b'    <DeviceMacId>0x0123456789abcdef</DeviceMacId>'
            b'</DeviceInfo>',
        b'<Command><Name>get_meter_list</Name></Command>': b'',
    }
    async with mock_pty_device(responses) as mock_device:
        async with RAVEnSerialDevice(mock_device) as dut:
            with pytest.raises(asyncio.TimeoutError):
                await asyncio.wait_for(dut.get_meter_list(), 0.05)
            assert await dut.get_device_info()


@pytest.mark.asyncio
async def test_serial_abort():
    """Verify behavior of device abort."""
    async with mock_pty_device({}) as mock_device:
        async with RAVEnSerialDevice(mock_device) as dut:
            await dut.abort()
        await dut.abort()



# Copyright 2022 Scott K Logan
# Licensed under the Apache License, Version 2.0

import asyncio
from collections.abc import AsyncIterator
from collections.abc import Awaitable
from collections.abc import MutableMapping
from contextlib import asynccontextmanager
import os
import sys
from typing import BinaryIO
from typing import Optional
from typing import SupportsIndex
from typing import TypeVar
from typing import Union

import serial_asyncio_fast


_T = TypeVar('_T')


class _NoPopList(list[_T]):

    def pop(self, index: SupportsIndex = -1) -> _T:
        return self[index]


_ResponseMapping = MutableMapping[bytes, Optional[Union[list[bytes], bytes]]]

DEFAULT_RESPONSES: _ResponseMapping = {
    b'<Command>'
    b'<Name>get_current_price</Name>'
    b'<MeterMacId>0xFEDCBA9876543210</MeterMacId>'
    b'</Command>': _NoPopList((
        b'<PriceCluster>'
        b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
        b'    <MeterMacId>0xFEDCBA9876543210</MeterMacId>'
        b'    <TimeStamp>0x29bd58a7</TimeStamp>'
        b'    <Price>0xc7</Price>'
        b'    <Currency>0x348</Currency>'
        b'    <TrailingDigits>0x03</TrailingDigits>'
        b'    <Tier>0x08</Tier>'
        b'    <TierLabel>Set by User</TierLabel>'
        b'    <RateLabel>Set by User</RateLabel>'
        b'</PriceCluster>',
    )),
    b'<Command>'
    b'<Name>get_current_summation_delivered</Name>'
    b'<MeterMacId>0xFEDCBA9876543210</MeterMacId>'
    b'</Command>': _NoPopList((
        b'<CurrentSummationDelivered>'
        b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
        b'    <MeterMacId>0xFEDCBA9876543210</MeterMacId>'
        b'    <TimeStamp>0x29bd58a7</TimeStamp>'
        b'    <SummationDelivered>0x00000010</SummationDelivered>'
        b'    <SummationReceived>0x00000008</SummationReceived>'
        b'    <Multiplier>0x00000004</Multiplier>'
        b'    <Divisor>0x00000002</Divisor>'
        b'    <DigitsRight>0x02</DigitsRight>'
        b'    <DigitsLeft>0x04</DigitsLeft>'
        b'    <SuppressLeadingZero>N</SuppressLeadingZero>'
        b'</CurrentSummationDelivered>',
    )),
    b'<Command>'
    b'<Name>get_device_info</Name>'
    b'</Command>': _NoPopList((
        b'<DeviceInfo>'
        b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
        b'    <InstallCode>0xABCDEF0123456789</InstallCode>'
        b'    <LinkKey>0xABCDEF0123456789ABCDEF0123456789</LinkKey>'
        b'    <FWVersion>1.21g</FWVersion>'
        b'    <HWVersion>5.55 rev 2</HWVersion>'
        b'    <ImageType>Mocked</ImageType>'
        b'    <Manufacturer>aioraven</Manufacturer>'
        b'    <ModelId>mock device</ModelId>'
        b'    <DateCode>20220101a0000042</DateCode>'
        b'</DeviceInfo>',
    )),
    b'<Command>'
    b'<Name>get_instantaneous_demand</Name>'
    b'<MeterMacId>0xFEDCBA9876543210</MeterMacId>'
    b'</Command>': _NoPopList((
        b'<InstantaneousDemand>'
        b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
        b'    <MeterMacId>0xFEDCBA9876543210</MeterMacId>'
        b'    <TimeStamp>0x29bd58a7</TimeStamp>'
        b'    <Demand>0x00000010</Demand>'
        b'    <Multiplier>0x00000004</Multiplier>'
        b'    <Divisor>0x00000002</Divisor>'
        b'    <DigitsRight>0x02</DigitsRight>'
        b'    <DigitsLeft>0x04</DigitsLeft>'
        b'    <SuppressLeadingZero>Y</SuppressLeadingZero>'
        b'</InstantaneousDemand>',
    )),
    b'<Command>'
    b'<Name>get_meter_info</Name>'
    b'<MeterMacId>0xFEDCBA9876543210</MeterMacId>'
    b'</Command>': _NoPopList((
        b'<MeterInfo>'
        b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
        b'    <MeterMacId>0xFEDCBA9876543210</MeterMacId>'
        b'    <MeterType>electric</MeterType>'
        b'    <NickName>House</NickName>'
        b'    <Account>8675309</Account>'
        b'    <Auth>p@ssw0rd</Auth>'
        b'    <Host>Example, Inc.</Host>'
        b'    <Enabled>Y</Enabled>'
        b'</MeterInfo>',
    )),
    b'<Command><Name>get_meter_list</Name></Command>': _NoPopList((
        b'<MeterList>'
        b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
        b'    <MeterMacId>0xFEDCBA9876543210</MeterMacId>'
        b'</MeterList>',
    )),
    b'<Command><Name>get_network_info</Name></Command>': _NoPopList((
        b'<NetworkInfo>'
        b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
        b'    <CoordMacId>0xFEDCBA9876543210</CoordMacId>'
        b'    <Status>Connected</Status>'
        b'    <Description>Network is operational</Description>'
        b'    <StatusCode>0x42</StatusCode>'
        b'    <ExtPanId>0x9876543210ABCDEF</ExtPanId>'
        b'    <Channel>24</Channel>'
        b'    <ShortAddr>0x5678</ShortAddr>'
        b'    <LinkStrength>0x24</LinkStrength>'
        b'</NetworkInfo>',
    )),
}


async def _device_loop(
    reader: asyncio.StreamReader,
    writer: asyncio.StreamWriter,
    responses: _ResponseMapping
) -> None:
    buffer = b''
    while True:
        try:
            try:
                value = await reader.read(1)
            except asyncio.CancelledError:
                break
            if not value:
                break
            buffer += value
            buffer = buffer.lstrip()
            for k, v in responses.items():
                if buffer.startswith(k):
                    if isinstance(v, list):
                        first = v.pop(0)
                        if first is not None:
                            writer.write(first)
                        if not v:
                            del responses[k]
                    else:
                        if v is not None:
                            writer.write(v)
                        del responses[k]
                    buffer = buffer[len(k):]
                    break
        except asyncio.CancelledError:
            if not writer.can_write_eof():
                break
            writer.write_eof()
            await writer.drain()
    writer.close()


@asynccontextmanager
async def mock_pty_device(
    responses: Optional[_ResponseMapping] = None,
    initial_buffer: Optional[bytes] = None
) -> AsyncIterator[str]:
    if responses is None:
        responses = DEFAULT_RESPONSES

    import pty

    server, client = pty.openpty()

    server_in = os.fdopen(server, 'rb', 0, closefd=False)
    server_out = os.fdopen(server, 'wb', 0, closefd=False)

    reader, writer, read_transport = await connect_pipes(server_in, server_out)

    if initial_buffer is not None:
        writer.write(initial_buffer)
    task = asyncio.create_task(_device_loop(reader, writer, responses))

    yield os.ttyname(client)

    task.cancel()

    try:
        await task
    except asyncio.CancelledError:
        pass

    read_transport.close()
    writer.close()

    server_in.close()
    server_out.close()
    os.close(server)
    os.close(client)


@asynccontextmanager
async def mock_device(
    responses: Optional[_ResponseMapping] = None,
    initial_buffer: Optional[bytes] = None
) -> AsyncIterator[None]:
    """
    Create a mock device at a TCP endpoint.

    This function creates a context-managed TCP endpoint which echoes verbatim
    responses given verbatim requests.

    :param dict responses: A mapping of request strings to responses.
    :param bytes initial_buffer: Content to initialize the response buffer.

    :returns: A tuple including the host and port of the TCP endpoint.
    """
    if responses is None:
        responses = DEFAULT_RESPONSES

    connections = []

    def client_connected(
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter
    ) -> Awaitable[None]:
        if initial_buffer is not None:
            writer.write(initial_buffer)
        task = asyncio.create_task(_device_loop(reader, writer, responses))
        connections.append(task)
        return asyncio.wait_for(task, None)

    server = await asyncio.start_server(client_connected, host='127.0.0.1')
    try:
        yield server.sockets[0].getsockname()
    finally:
        server.close()
    if not connections:
        return

    _, pending = await asyncio.wait(connections, timeout=0.1)
    for task in pending:
        task.cancel()
    if pending:
        await asyncio.wait(pending)


async def connect_pipes(
    in_pipe: BinaryIO,
    out_pipe: BinaryIO,
    loop: Optional[asyncio.AbstractEventLoop] = None,
) -> tuple[asyncio.StreamReader, asyncio.StreamWriter, asyncio.ReadTransport]:
    if loop is None:
        loop = asyncio.get_event_loop()
    reader = asyncio.StreamReader()
    protocol = asyncio.StreamReaderProtocol(reader)

    read_transport, _ = await loop.connect_read_pipe(
        lambda: protocol, in_pipe)
    write_transport, write_protocol = await loop.connect_write_pipe(
        asyncio.streams.FlowControlMixin, out_pipe)
    writer = asyncio.StreamWriter(
        write_transport, write_protocol, reader, loop)
    return reader, writer, read_transport


async def main(argv: list[str] = sys.argv) -> Optional[int]:
    if len(argv) == 1:
        reader, writer, read_transport = await connect_pipes(
            sys.stdin.buffer, sys.stdout.buffer)
    elif len(argv) == 2:
        read_transport = None
        reader, writer = await serial_asyncio_fast.open_serial_connection(
            url=argv[1])
    else:
        print(f'Usage: {argv[0]} [DEVICE_PATH]', file=sys.stderr)
        return 1
    await _device_loop(reader, writer, DEFAULT_RESPONSES)
    if read_transport is not None:
        read_transport.close()
    return 0


if __name__ == '__main__':
    sys.exit(asyncio.run(main()))



# Copyright 2022 Scott K Logan
# Licensed under the Apache License, Version 2.0

from asyncio.events import AbstractEventLoop
from asyncio.events import get_event_loop
from asyncio.futures import Future
from asyncio.protocols import Protocol
from asyncio.transports import BaseTransport
from typing import Any
from typing import Optional
import warnings
import xml.etree.ElementTree as Et

from aioraven.device import RAVEnWarning
from aioraven.device import UnknownRAVEnCommandWarning
from aioraven.reader import RAVEnReader


class RAVEnReaderProtocol(Protocol):
    """Deserialize data fragments from a RAVEn device."""

    _closed: Future[None]
    _reader: Optional[RAVEnReader]
    _stash: bytes

    def __init__(
        self,
        reader: RAVEnReader,
        loop: Optional[AbstractEventLoop] = None,
    ) -> None:
        """
        Construct a RAVEnRaderProtocol.

        :param reader: The `RAVEnReader` instance to which deserialized
          fragments are passed.
        :param loop: The event loop instance to use.
        """
        if loop is None:
            self._loop = get_event_loop()
        else:
            self._loop = loop
        self._reader = reader
        self._closed = self._loop.create_future()

    def _reset(self) -> None:
        self._parser = Et.XMLPullParser(events=('end',))
        self._parser.feed(b'<?xml version="1.0" encoding="ASCII"?><root>')
        self._stash = b''

    def _get_close_waiter(self, stream: Any) -> Future[None]:
        return self._closed

    def connection_made(self, transport: BaseTransport) -> None:
        self._reset()

    def connection_lost(self, exc: Optional[Exception]) -> None:
        if self._reader is not None:
            if exc is None:
                self._reader.feed_eof()
            else:
                self._reader.set_exception(exc)
        if not self._closed.done():
            if exc is None:
                self._closed.set_result(None)
            else:
                self._closed.set_exception(exc)
        self._reader = None

    def data_received(self, data: bytes) -> None:
        if not self._reader:
            return

        self._stash += data
        if b'>' not in data:
            return

        self._parser.feed(self._stash)
        self._stash = b''

        events = self._parser.read_events()
        while True:
            try:
                _, element = next(events)
            except StopIteration:
                return
            except Et.ParseError as err:
                self._reader.set_exception(err)
                self._reset()
            else:
                if element.tag == 'Warning':
                    try:
                        e = next(iter(element), None)
                        if e is not None and e.text is not None:
                            if e.text == UnknownRAVEnCommandWarning.MESSAGE:
                                warnings.warn(UnknownRAVEnCommandWarning())
                            else:
                                warnings.warn(RAVEnWarning(e.text))
                        else:
                            warnings.warn(RAVEnWarning('Unknown warning'))
                    except Warning as err:
                        self._reader.set_exception(err)
                else:
                    self._reader.feed_element(element)

    def eof_received(self) -> None:
        if not self._reader:
            return

        self._parser.feed(b'</root>')

        try:
            self._parser.close()
        except Et.ParseError as err:
            self._reader.set_exception(err)
        finally:
            self._reader.feed_eof()

    def __del__(self) -> None:
        try:
            closed = self._closed
        except AttributeError:
            pass
        else:
            if closed.done() and not closed.cancelled():
                closed.exception()



# Copyright 2022 Scott K Logan
# Licensed under the Apache License, Version 2.0

from asyncio.events import AbstractEventLoop
from asyncio.events import get_event_loop
from asyncio.futures import Future
from typing import Dict
from typing import List
from typing import Optional
import xml.etree.ElementTree as Et

from aioraven.data import RAVEnData


class RAVEnReader:
    """Read stanzas from a RAVEn device."""

    _exception: Optional[Exception]
    _waiters: Dict[Optional[str], List[Future[Optional[RAVEnData]]]]

    def __init__(self, loop: Optional[AbstractEventLoop] = None) -> None:
        """
        Construct a RAVEnReader.

        :param loop: The event loop instance to use.
        """
        if loop is None:
            self._loop = get_event_loop()
        else:
            self._loop = loop
        self._eof = False
        self._waiters = {None: []}
        self._exception = None

    def __repr__(self) -> str:
        info = [self.__class__.__name__]
        if self._eof:
            info.append('eof')
        if self._exception:
            info.append('e=%r' % self._exception)
        num_waiters = sum(len(w) for w in self._waiters.values())
        if num_waiters:
            info.append('w=%d' % num_waiters)
        return '<%s>' % ' '.join(info)

    def exception(self) -> Optional[Exception]:
        return self._exception

    def set_exception(self, exc: Exception) -> None:
        # Only ParseError is recoverable
        if not isinstance(exc, Et.ParseError):
            self._exception = exc
        for waiters in self._waiters.values():
            while waiters:
                waiter = waiters.pop(0)
                if not waiter.cancelled():
                    waiter.set_exception(exc)

    def feed_eof(self) -> None:
        self._eof = True
        for waiters in self._waiters.values():
            while waiters:
                waiter = waiters.pop(0)
                if not waiter.cancelled():
                    waiter.set_result(None)

    def feed_element(self, data: Et.Element) -> None:
        self._waiters.setdefault(data.tag, [])
        waiters = self._waiters.get(data.tag, [])
        res: RAVEnData = {}
        for e in data:
            if e.tag in res:
                orig = res[e.tag]
                if not isinstance(orig, list):
                    res[e.tag] = [orig, e.text]
                else:
                    orig.append(e.text)
            else:
                res[e.tag] = e.text
        while waiters:
            waiter = waiters.pop(0)
            if not waiter.cancelled():
                waiter.set_result(res)
                break
        else:
            waiters = self._waiters[None]
            if waiters:
                waiters.pop(0).set_result(res)

    async def read_tag(
        self,
        tag: Optional[str] = None,
    ) -> Optional[RAVEnData]:
        if self._eof:
            return None
        if self._exception is not None:
            raise self._exception
        self._waiters.setdefault(tag, [])
        waiter: Future[Optional[RAVEnData]] = self._loop.create_future()
        self._waiters[tag].append(waiter)
        return await waiter



# Copyright 2022 Scott K Logan
# Licensed under the Apache License, Version 2.0

from datetime import datetime
from typing import Dict
from typing import Optional

from aioraven.data import ConnectionState
from aioraven.data import convert_bool
from aioraven.data import convert_currency
from aioraven.data import convert_date_code
from aioraven.data import convert_datetime
from aioraven.data import convert_float_formatted
from aioraven.data import convert_hex_to_bytes
from aioraven.data import convert_int
from aioraven.data import convert_price
from aioraven.data import convert_str
from aioraven.data import convert_timedelta
from aioraven.data import CurrentPeriodUsage
from aioraven.data import CurrentSummationDelivered
from aioraven.data import DataStatus
from aioraven.data import DeviceInfo
from aioraven.data import InstantaneousDemand
from aioraven.data import IntervalChannel
from aioraven.data import IntervalPeriod
from aioraven.data import LastPeriodUsage
from aioraven.data import MessageCluster
from aioraven.data import MessageQueue
from aioraven.data import MeterInfo
from aioraven.data import MeterList
from aioraven.data import MeterType
from aioraven.data import NetworkInfo
from aioraven.data import PriceCluster
from aioraven.data import ProfileData
from aioraven.data import RAVEnData
from aioraven.data import ScheduledEvent
from aioraven.data import ScheduleInfo
from aioraven.data import TimeCluster


class RAVEnError(IOError):
    """Base class for RAVEn related exceptions."""


class RAVEnConnectionError(RAVEnError):
    """Communication error with a RAVEn device."""


class RAVEnNotOpenError(RAVEnError):
    """The RAVEn device is not open."""


class RAVEnWarning(Warning):
    """The RAVEn device has generated a warning message."""

    def __init__(self, message: str) -> None:
        """
        Construct a RAVEnWarning.

        :param message: The reason for which the device is warning us.
        """
        super().__init__('Warning from RAVEn device: ' + message)


class UnknownRAVEnCommandWarning(RAVEnWarning):
    """A recently executed command is not supported by the RAVEn device."""

    MESSAGE = 'Unknown command'

    def __init__(self) -> None:
        """Construct a UnknownRAVEnCommandWarning."""
        super().__init__(self.MESSAGE)


class RAVEnBaseDevice:
    """RAVEn device command implementation."""

    async def _query(
        self,
        cmd_name: str,
        res_name: Optional[str] = None,
        args: Optional[Dict[str, str]] = None,
    ) -> Optional[RAVEnData]:
        raise NotImplementedError(
            "Derived class must implement '_query' function")

    async def close_current_period(
        self,
        meter: Optional[bytes] = None
    ) -> None:
        args = {}
        if meter is not None:
            args['MeterMacId'] = f'0x{meter.hex().upper()}'
        await self._query('close_current_period', None, args)

    async def confirm_message(
        self,
        msg_id: int,
        meter: Optional[bytes] = None,
    ) -> None:
        args = {
            'Id': f'0x{msg_id:08X}',
        }
        if meter is not None:
            args['MeterMacId'] = f'0x{meter.hex().upper()}'
        await self._query('confirm_message', None, args)

    async def factory_reset(self) -> None:
        await self._query('factory_reset')

    async def get_current_period_usage(
        self,
        meter: Optional[bytes] = None,
    ) -> Optional[CurrentPeriodUsage]:
        args = {}
        if meter is not None:
            args['MeterMacId'] = f'0x{meter.hex().upper()}'
        raw = await self._query(
            'get_current_period_usage', 'CurrentPeriodUsage', args)
        if not raw:
            return None
        return CurrentPeriodUsage(
            device_mac_id=convert_hex_to_bytes(
                convert_str(raw.get('DeviceMacId'))),
            meter_mac_id=convert_hex_to_bytes(
                convert_str(raw.get('MeterMacId'))),
            time_stamp=convert_datetime(
                convert_str(raw.get('TimeStamp')), True),
            current_usage=convert_float_formatted(
                convert_str(raw.get('CurrentUsage')),
                convert_str(raw.get('Multiplier')),
                convert_str(raw.get('Divisor')),
                convert_str(raw.get('DigitsRight')),
                convert_str(raw.get('DigitsLeft')),
                convert_str(raw.get('SuppressLeadingZero'))),
            start_date=convert_datetime(
                convert_str(raw.get('StartDate')), True))

    async def get_current_price(
        self,
        *,
        meter: Optional[bytes] = None,
        refresh: Optional[bool] = None,
    ) -> Optional[PriceCluster]:
        args = {}
        if meter is not None:
            args['MeterMacId'] = f'0x{meter.hex().upper()}'
        if refresh is not None:
            args['Refresh'] = 'Y' if refresh else 'N'
        raw = await self._query('get_current_price', 'PriceCluster', args)
        if not raw:
            return None
        currency = convert_currency(convert_str(raw.get('Currency')))
        return PriceCluster(
            device_mac_id=convert_hex_to_bytes(
                convert_str(raw.get('DeviceMacId'))),
            meter_mac_id=convert_hex_to_bytes(
                convert_str(raw.get('MeterMacId'))),
            time_stamp=convert_datetime(
                convert_str(raw.get('TimeStamp')), True),
            price=convert_price(
                convert_str(raw.get('Price')),
                convert_str(raw.get('TrailingDigits')),
                currency.exponent if currency else None),
            currency=currency,
            tier=convert_int(convert_str(raw.get('Tier'))),
            tier_label=convert_str(raw.get('TierLabel')),
            rate_label=convert_str(raw.get('RateLabel')))

    async def get_current_summation_delivered(
        self,
        *,
        meter: Optional[bytes] = None,
        refresh: Optional[bool] = None,
    ) -> Optional[CurrentSummationDelivered]:
        args = {}
        if meter is not None:
            args['MeterMacId'] = f'0x{meter.hex().upper()}'
        if refresh is not None:
            args['Refresh'] = 'Y' if refresh else 'N'
        raw = await self._query(
            'get_current_summation_delivered', 'CurrentSummationDelivered',
            args)
        if not raw:
            return None
        return CurrentSummationDelivered(
            device_mac_id=convert_hex_to_bytes(
                convert_str(raw.get('DeviceMacId'))),
            meter_mac_id=convert_hex_to_bytes(
                convert_str(raw.get('MeterMacId'))),
            time_stamp=convert_datetime(
                convert_str(raw.get('TimeStamp')), True),
            summation_delivered=convert_float_formatted(
                convert_str(raw.get('SummationDelivered')),
                convert_str(raw.get('Multiplier')),
                convert_str(raw.get('Divisor')),
                convert_str(raw.get('DigitsRight')),
                convert_str(raw.get('DigitsLeft')),
                convert_str(raw.get('SuppressLeadingZero'))),
            summation_received=convert_float_formatted(
                convert_str(raw.get('SummationReceived')),
                convert_str(raw.get('Multiplier')),
                convert_str(raw.get('Divisor')),
                convert_str(raw.get('DigitsRight')),
                convert_str(raw.get('DigitsLeft')),
                convert_str(raw.get('SuppressLeadingZero'))))

    async def get_device_info(self) -> Optional[DeviceInfo]:
        raw = await self._query('get_device_info', 'DeviceInfo')
        if not raw:
            return None
        return DeviceInfo(
            device_mac_id=convert_hex_to_bytes(
                convert_str(raw.get('DeviceMacId'))),
            install_code=convert_hex_to_bytes(
                convert_str(raw.get('InstallCode'))),
            link_key=convert_hex_to_bytes(
                convert_str(raw.get('LinkKey'))),
            fw_version=convert_str(raw.get('FWVersion')),
            hw_version=convert_str(raw.get('HWVersion')),
            image_type=convert_str(raw.get('ImageType')),
            manufacturer=convert_str(raw.get('Manufacturer')),
            model_id=convert_str(raw.get('ModelId')),
            date_code=convert_date_code(convert_str(raw.get('DateCode'))))

    async def get_instantaneous_demand(
        self,
        *,
        meter: Optional[bytes] = None,
        refresh: Optional[bool] = None,
    ) -> Optional[InstantaneousDemand]:
        args = {}
        if meter is not None:
            args['MeterMacId'] = f'0x{meter.hex().upper()}'
        if refresh is not None:
            args['Refresh'] = 'Y' if refresh else 'N'
        raw = await self._query(
            'get_instantaneous_demand', 'InstantaneousDemand', args)
        if not raw:
            return None
        return InstantaneousDemand(
            device_mac_id=convert_hex_to_bytes(
                convert_str(raw.get('DeviceMacId'))),
            meter_mac_id=convert_hex_to_bytes(
                convert_str(raw.get('MeterMacId'))),
            time_stamp=convert_datetime(
                convert_str(raw.get('TimeStamp')), True),
            demand=convert_float_formatted(
                convert_str(raw.get('Demand')),
                convert_str(raw.get('Multiplier')),
                convert_str(raw.get('Divisor')),
                convert_str(raw.get('DigitsRight')),
                convert_str(raw.get('DigitsLeft')),
                convert_str(raw.get('SuppressLeadingZero'))))

    async def get_last_period_usage(
        self,
        *,
        meter: Optional[bytes] = None,
    ) -> Optional[LastPeriodUsage]:
        # TODO(cottsay): This command has not been tested
        args = {}
        if meter is not None:
            args['MeterMacId'] = f'0x{meter.hex().upper()}'
        raw = await self._query(
            'get_last_period_usage', 'LastPeriodUsage', args)
        if not raw:
            return None
        return LastPeriodUsage(
            device_mac_id=convert_hex_to_bytes(
                convert_str(raw.get('DeviceMacId'))),
            meter_mac_id=convert_hex_to_bytes(
                convert_str(raw.get('MeterMacId'))),
            last_usage=convert_float_formatted(
                convert_str(raw.get('LastUsage')),
                convert_str(raw.get('Multiplier')),
                convert_str(raw.get('Divisor')),
                convert_str(raw.get('DigitsRight')),
                convert_str(raw.get('DigitsLeft')),
                convert_str(raw.get('SuppressLeadingZero'))),
            start_date=convert_datetime(
                convert_str(raw.get('StartDate')), True),
            end_date=convert_datetime(
                convert_str(raw.get('EndDate')), True))

    async def get_message(
        self,
        *,
        meter: Optional[bytes] = None,
        refresh: Optional[bool] = None,
    ) -> Optional[MessageCluster]:
        args = {}
        if meter is not None:
            args['MeterMacId'] = f'0x{meter.hex().upper()}'
        if refresh is not None:
            args['Refresh'] = 'Y' if refresh else 'N'
        raw = await self._query('get_message', 'MessageCluster', args)
        if not raw:
            return None
        queue = convert_str(raw.get('Queue'))
        return MessageCluster(
            device_mac_id=convert_hex_to_bytes(
                convert_str(raw.get('DeviceMacId'))),
            meter_mac_id=convert_hex_to_bytes(
                convert_str(raw.get('MeterMacId'))),
            time_stamp=convert_datetime(
                convert_str(raw.get('TimeStamp')), True),
            message_id=convert_hex_to_bytes(
                convert_str(raw.get('Id'))),
            text=convert_str(
                convert_str(raw.get('Text'))),
            confirmation_required=convert_bool(
                convert_str(raw.get('ConfirmationRequired'))),
            confirmed=convert_bool(
                convert_str(raw.get('Confirmed'))),
            queue=MessageQueue(queue) if queue else None)

    async def get_meter_info(
        self,
        *,
        meter: Optional[bytes] = None,
    ) -> Optional[MeterInfo]:
        args = {}
        if meter is not None:
            args['MeterMacId'] = f'0x{meter.hex().upper()}'
        raw = await self._query('get_meter_info', 'MeterInfo', args)
        if not raw:
            return None
        # TODO(cottsay): Got unexpected numeric '0x0000' instead of 'electric'
        raw_type = convert_str(raw.get('MeterType'))
        return MeterInfo(
            device_mac_id=convert_hex_to_bytes(
                convert_str(raw.get('DeviceMacId'))),
            meter_mac_id=convert_hex_to_bytes(
                convert_str(raw.get('MeterMacId'))),
            meter_type=MeterType(raw_type) if raw_type else None,
            nick_name=convert_str(raw.get('NickName')),
            account=convert_str(raw.get('Account')),
            auth=convert_str(raw.get('Auth')),
            host=convert_str(raw.get('Host')),
            enabled=convert_bool(convert_str(raw.get('Enabled'))))

    async def get_meter_list(self) -> Optional[MeterList]:
        raw = await self._query('get_meter_list', 'MeterList')
        if not raw:
            return None
        raw_list = raw.get('MeterMacId') or []
        if not isinstance(raw_list, list):
            raw_list = [raw_list]
        converted_list = [convert_hex_to_bytes(m) for m in raw_list]
        return MeterList(
            device_mac_id=convert_hex_to_bytes(
                convert_str(raw.get('DeviceMacId'))),
            meter_mac_ids=[m for m in converted_list if m is not None])

    async def get_network_info(self) -> Optional[NetworkInfo]:
        raw = await self._query('get_network_info', 'NetworkInfo')
        if not raw:
            return None
        status = convert_str(raw.get('Status'))
        return NetworkInfo(
            device_mac_id=convert_hex_to_bytes(
                convert_str(raw.get('DeviceMacId'))),
            coord_mac_id=convert_hex_to_bytes(
                convert_str(raw.get('CoordMacId'))),
            status=ConnectionState(status) if status else None,
            description=convert_str(raw.get('Description')),
            status_code=convert_hex_to_bytes(
                convert_str(raw.get('StatusCode'))),
            ext_pan_id=convert_hex_to_bytes(convert_str(raw.get('ExtPanId'))),
            channel=convert_int(convert_str(raw.get('Channel'))),
            short_addr=convert_hex_to_bytes(convert_str(raw.get('ShortAddr'))),
            link_strength=convert_int(convert_str(raw.get('LinkStrength'))))

    async def get_profile_data(
        self,
        count: int,
        end: datetime,
        channel: IntervalChannel,
        *,
        meter: Optional[bytes] = None,
    ) -> Optional[ProfileData]:
        # TODO(cottsay): This command has not been tested
        args = {
            'NumberOfPeriods': f'0x{count:02X}',
            'EndTime': str(end),
            'IntervalChannel': f'{channel}',
        }
        if meter is not None:
            args['MeterMacId'] = f'0x{meter.hex().upper()}'
        raw = await self._query(
            'get_profile_data', 'ProfileData', args)
        if not raw:
            return None
        status = convert_int(convert_str(raw.get('Status')))
        period = convert_str(raw.get('ProfileIntervalPeriod'))
        return ProfileData(
            device_mac_id=convert_hex_to_bytes(
                convert_str(raw.get('DeviceMacId'))),
            meter_mac_id=convert_hex_to_bytes(
                convert_str(raw.get('MeterMacId'))),
            end_time=convert_datetime(
                convert_str(raw.get('EndTime')), True),
            status=DataStatus(status) if status is not None else None,
            profile_interval_period=IntervalPeriod(
                int(period)
            ) if period is not None else None)
        # TODO(cottsay): Get the IntervalData

    async def get_schedule(
        self,
        *,
        meter: Optional[bytes] = None,
        event: Optional[ScheduledEvent] = None,
    ) -> Optional[ScheduleInfo]:
        args = {}
        if meter is not None:
            args['MeterMacId'] = f'0x{meter.hex().upper()}'
        if event is not None:
            args['Event'] = str(event)
        raw = await self._query('get_schedule', 'ScheduleInfo', args)
        if not raw:
            return None
        scheduled_event = convert_str(raw.get('Event'))
        return ScheduleInfo(
            device_mac_id=convert_hex_to_bytes(
                convert_str(raw.get('DeviceMacId'))),
            meter_mac_id=convert_hex_to_bytes(
                convert_str(raw.get('MeterMacId'))),
            event=ScheduledEvent(scheduled_event) if scheduled_event else None,
            frequency=convert_timedelta(
                convert_str(raw.get('Frequency'))),
            enabled=convert_bool(
                convert_str(raw.get('Enabled'))))

    async def get_time(
        self,
        *,
        meter: Optional[bytes] = None,
        refresh: Optional[bool] = None,
    ) -> Optional[TimeCluster]:
        args = {}
        if meter is not None:
            args['MeterMacId'] = f'0x{meter.hex().upper()}'
        if refresh is not None:
            args['Refresh'] = 'Y' if refresh else 'N'
        raw = await self._query('get_time', 'TimeCluster', args)
        if not raw:
            return None
        return TimeCluster(
            device_mac_id=convert_hex_to_bytes(
                convert_str(raw.get('DeviceMacId'))),
            meter_mac_id=convert_hex_to_bytes(
                convert_str(raw.get('MeterMacId'))),
            utc_time=convert_datetime(
                convert_str(raw.get('UTCTime')), True),
            local_time=convert_datetime(
                convert_str(raw.get('LocalTime'))))

    async def initialize(self) -> None:
        # TODO(cottsay): This command has not been tested
        await self._query('initialize')

    async def restart(self) -> None:
        await self._query('restart')

    async def set_current_price(
        self,
        price: float,
        *,
        meter: Optional[bytes] = None,
    ) -> None:
        args = {
            'Price': str(price),
            'TrailingDigits': str(0),
        }
        # TODO(cottsay): Handle TrailingDigits
        if meter is not None:
            args['MeterMacId'] = str(meter)
        await self._query('set_current_price', None, args)

    async def set_fast_poll(
        self,
        frequency: int,
        duration: int,
        *,
        meter: Optional[bytes] = None,
    ) -> None:
        args = {
            'Frequency': str(frequency),
            'Duration': str(duration),
        }
        if meter is not None:
            args['MeterMacId'] = f'0x{meter.hex().upper()}'
        await self._query('set_fast_poll', None, args)

    async def set_meter_info(
        self,
        *,
        meter: Optional[bytes] = None,
        nick: Optional[str] = None,
        account: Optional[str] = None,
        auth: Optional[str] = None,
        host: Optional[str] = None,
        enabled: Optional[bool] = None,
    ) -> None:
        args = {}
        if meter is not None:
            args['MeterMacId'] = f'0x{meter.hex().upper()}'
        if nick is not None:
            args['NickName'] = str(nick)
        if account is not None:
            args['Account'] = str(account)
        if auth is not None:
            args['Auth'] = str(auth)
        if host is not None:
            args['host'] = str(host)
        if enabled is not None:
            args['enabled'] = 'Y' if enabled else 'N'
        await self._query('set_meter_info', None, args)

    async def set_schedule(
        self,
        event: ScheduledEvent,
        frequency: int,
        enabled: bool,
        *,
        meter: Optional[bytes] = None,
    ) -> None:
        args = {
            'Event': str(event),
            'Frequency': str(frequency),
            'Enabled': 'Y' if enabled else 'N',
        }
        if meter is not None:
            args['MeterMacId'] = f'0x{meter.hex().upper()}'
        await self._query('set_schedule', None, args)

    async def set_schedule_default(
        self,
        *,
        meter: Optional[bytes] = None,
        event: Optional[ScheduledEvent] = None,
    ) -> None:
        args = {}
        if meter is not None:
            args['MeterMacId'] = f'0x{meter.hex().upper()}'
        if event is not None:
            args['Event'] = str(event)
        await self._query('set_schedule_default', None, args)



# Copyright 2022 Scott K Logan
# Licensed under the Apache License, Version 2.0

import asyncio

from aioraven.data import MeterList
from aioraven.device import RAVEnConnectionError
from aioraven.device import RAVEnNotOpenError
from aioraven.streams import open_connection
from aioraven.streams import RAVEnNetworkDevice
import pytest

from .mock_device import mock_device


@pytest.mark.asyncio
async def test_tcp_data():
    """Verify simple device query behavior."""
    responses = {
        b'<Command><Name>get_meter_list</Name></Command>':
            b'<MeterList>'
            b'    <DeviceMacId>0x0123456789abcdef</DeviceMacId>'
            b'</MeterList>',
    }

    async with mock_device(responses) as (host, port):
        async with RAVEnNetworkDevice(host, port) as dut:
            actual = await dut.get_meter_list()

    assert actual == MeterList(
        device_mac_id=bytes.fromhex('0123456789ABCDEF'),
        meter_mac_ids=[])


@pytest.mark.asyncio
async def test_tcp_disconnect():
    """Verify behavior when a device is unexpectedly disconnected."""
    responses = {
        b'<Command><Name>get_meter_list</Name></Command>':
            b'<MeterList>'
            b'    <DeviceMacId>0x0123456789abcdef</DeviceMacId>'
            b'</MeterList>',
    }

    async with mock_device(responses) as (host, port):
        dut = RAVEnNetworkDevice(host, port)
        await dut.open()
        assert await dut.get_meter_list()
    async with dut:
        assert not await dut.get_meter_list()
        assert not await dut.get_meter_list()


@pytest.mark.asyncio
async def test_tcp_incomplete():
    """Verify behavior when a partial fragment is received."""
    responses = {
        b'<Command><Name>get_meter_list</Name></Command>':
            b'<MeterList>'
            b'    <DeviceMacId>0x0123456789abcdef</DeviceMacId>'
            b'</MeterList>'
            b'<DeviceInfo>'
    }

    async with mock_device(responses) as (host, port):
        dut = RAVEnNetworkDevice(host, port)
        await dut.open()
        assert await dut.get_meter_list()

        task = asyncio.create_task(dut.get_meter_list())
        await asyncio.wait((task,), timeout=0.05)
    async with dut:
        with pytest.raises(RAVEnConnectionError):
            await task
        assert not await dut.get_meter_list()


@pytest.mark.asyncio
async def test_tcp_not_open():
    """Verify behavior when reading from an unopened device."""
    async with mock_device() as (host, port):
        dut = RAVEnNetworkDevice(host, port)
        with pytest.raises(RAVEnNotOpenError):
            await dut.get_device_info()


@pytest.mark.asyncio
async def test_tcp_parse_error():
    """Verify behavior when invalid syntax is received."""
    responses = {
        b'<Command><Name>get_device_info</Name></Command>':
            b'</DeviceInfo>',
        b'<Command><Name>get_meter_list</Name></Command>':
            b'<MeterList>'
            b'    <DeviceMacId>0x0123456789abcdef</DeviceMacId>'
            b'</MeterList>',
    }

    async with mock_device(responses) as (host, port):
        async with RAVEnNetworkDevice(host, port) as dut:
            with pytest.raises(RAVEnConnectionError):
                await dut.get_device_info()
            actual = await dut.get_meter_list()

    assert actual == MeterList(
        device_mac_id=bytes.fromhex('0123456789ABCDEF'),
        meter_mac_ids=[])


@pytest.mark.asyncio
async def test_tcp_repr():
    """Verify representation of a network device."""
    async with mock_device() as (host, port):
        async with RAVEnNetworkDevice(host, port) as dut:
            assert 'RAVEnNetworkDevice' in str(dut)

        reader, writer = await open_connection(host, port)
        try:
            assert 'RAVEnReader' in str(reader)
            assert 'RAVEnWriter' in str(writer)
        finally:
            writer.close()


@pytest.mark.asyncio
async def test_tcp_timeout_recovery():
    """Verify device recovery after a timeout."""
    responses = {
        b'<Command><Name>get_device_info</Name></Command>':
            b'<DeviceInfo>'
            b'    <DeviceMacId>0x0123456789abcdef</DeviceMacId>'
            b'</DeviceInfo>',
        b'<Command><Name>get_meter_list</Name></Command>': b'',
    }
    async with mock_device(responses) as (host, port):
        async with RAVEnNetworkDevice(host, port) as dut:
            with pytest.raises(asyncio.TimeoutError):
                await asyncio.wait_for(dut.get_meter_list(), 0.05)
            assert await dut.get_device_info()


@pytest.mark.asyncio
async def test_tcp_abort():
    """Verify behavior of device abort."""
    async with mock_device({}) as (host, port):
        async with RAVEnNetworkDevice(host, port) as dut:
            await dut.abort()
        await dut.abort()



# Copyright 2022 Scott K Logan
# Licensed under the Apache License, Version 2.0

from asyncio.events import AbstractEventLoop
from asyncio.events import get_event_loop
from typing import Optional
from typing import Tuple

from aioraven.device import RAVEnConnectionError
from aioraven.protocols import RAVEnReaderProtocol
from aioraven.reader import RAVEnReader
from aioraven.streams import RAVEnStreamDevice
from aioraven.streams import RAVEnWriter
from serial import SerialException
from serial_asyncio_fast import create_serial_connection


async def open_serial_connection(
    url: str,
    *args: int,
    loop: Optional[AbstractEventLoop] = None,
    **kwargs: int,
) -> Tuple[RAVEnReader, RAVEnWriter]:
    """
    Establish a serial connection to a RAVEn device.

    Positional and keyword arguments are passed to
    `serial_asyncio_fast.create_serial_connection()`.

    :param loop: The event loop instance to use.
    """
    if loop is None:
        loop = get_event_loop()
    reader = RAVEnReader(loop=loop)
    protocol = RAVEnReaderProtocol(reader, loop=loop)
    kwargs.setdefault('baudrate', 115200)
    try:
        transport, _ = await create_serial_connection(
            loop, lambda: protocol, url, *args, **kwargs)
    except SerialException as ex:
        raise RAVEnConnectionError(f'{ex}') from ex
    writer = RAVEnWriter(transport, protocol)
    return reader, writer


class RAVEnSerialDevice(RAVEnStreamDevice):
    """A serial-connected RAVEn device."""

    def __init__(
        self,
        url: str,
        *args: int,
        loop: Optional[AbstractEventLoop] = None,
        **kwargs: int
    ) -> None:
        """
        Construct a RAVEnSerialDevice.

        Additional positional and keyword arguments are passed to
        `serial_asyncio_fast.create_serial_connection()`.

        :param url: The pyserial URL of the device to connect to.
        :param loop: The event loop instance to use.
        """
        self._url = url
        self._args = args
        self._loop = loop
        self._kwargs = kwargs

    def __repr__(self) -> str:
        info = [self.__class__.__name__]
        info.append('url=%s' % self._url)
        return '<%s>' % ' '.join(info)

    async def open(self) -> None:
        """Open the connection to the RAVEn device."""
        if self._reader or self._writer:
            return
        self._reader, self._writer = await open_serial_connection(
            self._url, *self._args, loop=self._loop, **self._kwargs)



# Copyright 2022 Scott K Logan
# Licensed under the Apache License, Version 2.0

from dataclasses import dataclass
from datetime import date
from datetime import datetime
from datetime import timedelta
from datetime import timezone
from enum import Enum
from enum import IntEnum
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

from iso4217 import Currency


RAVEnDatum = Union[Optional[str], List[Optional[str]]]
RAVEnData = Dict[str, RAVEnDatum]


def convert_str(raw: RAVEnDatum) -> Optional[str]:
    if raw is not None:
        if isinstance(raw, str):
            return raw
        raise ValueError(f'Expected a single value but got {len(raw)}')
    return None


def convert_bool(raw: Optional[str]) -> Optional[bool]:
    if raw is not None:
        if raw == 'Y':
            return True
        elif raw == 'N':
            return False
        raise ValueError(f"Invalid boolean format: '{raw}'")
    return None


def convert_currency(raw: Optional[str]) -> Optional[Currency]:
    if raw is not None:
        number = int(raw, 0)
        for currency in Currency:
            if currency.number == number:
                return currency
        raise ValueError(f"Invalid currency number: '{raw}'")
    return None


def convert_date(raw: Optional[str]) -> Optional[date]:
    if raw is not None:
        if len(raw) >= 8:
            return date(int(raw[:4]), int(raw[4:6]), int(raw[6:8]))
        raise ValueError(f"Invalid date format: '{raw}'")
    return None


def convert_date_code(raw: Optional[str]) -> Optional[Tuple[date, str]]:
    if raw is not None:
        if len(raw) >= 9:
            date = convert_date(raw[:8])
            if date is not None:
                return (date, raw[8:])
            return None
        raise ValueError(f"Invalid date code format: '{raw}'")
    return None


def convert_datetime(
    raw: Optional[str],
    utc: bool = False
) -> Optional[datetime]:
    if raw is not None and raw != '0xffffff':
        value = datetime(
            2000, 1, 1, 0, 0,
            tzinfo=timezone.utc if utc else None)
        return value + timedelta(seconds=int(raw, 0))
    return None


def convert_float(
    raw_value: Optional[str],
    raw_div: Optional[str]
) -> Optional[float]:
    if raw_value is not None:
        value = float(int(raw_value, 0))
        if raw_div is not None:
            return value / 10.0 ** int(raw_div, 0)
        return value
    return None


def convert_float_formatted(
    raw: Optional[str],
    mult: Optional[str] = None,
    div: Optional[str] = None,
    digits_right: Optional[str] = None,
    digits_left: Optional[str] = None,
    suppress_leading_zero: Optional[str] = None,
) -> Optional[str]:
    value_int = convert_int(raw, sign_extend=True)
    if value_int is not None:
        value = float(value_int)
        if mult is not None:
            value *= int(mult, 0)
        if div is not None:
            value /= int(div, 0)
        whole, frac = str(value).split('.', 1)
        if whole.startswith('-'):
            sign = '-'
            whole = whole[1:]
        else:
            sign = ''
        if digits_right is not None:
            places = int(digits_right, 0)
            difference = len(frac) - places
            if difference > 0:
                frac = str((int(frac) + 5 ** difference) // 10 ** difference)
                if len(frac) > places:
                    whole = str(int(whole) + 1)
                    frac = frac[1:]
            elif difference < 0:
                frac += '0' * (-difference)
        if digits_left is not None:
            whole = format(whole, f'0>{int(digits_left, 0)}')
        if suppress_leading_zero is not None:
            if suppress_leading_zero == 'Y':
                whole = whole.lstrip('0')
                if not whole:
                    whole = '0'
        return f'{sign}{whole}.{frac}'
    return None


def convert_hex_to_bytes(raw: Optional[str]) -> Optional[bytes]:
    if raw is not None:
        if raw.startswith('0x'):
            raw = raw[2:]
        return bytes.fromhex(raw)
    return None


def convert_int(
    raw: Optional[str],
    *,
    sign_extend: bool = False,
) -> Optional[int]:
    if raw is not None:
        value = int(raw, 0)
        if sign_extend:
            value = (value & 0x7FFFFFFF) - (value & 0x80000000)
        return value
    return None


def convert_price(
    raw_value: Optional[str],
    raw_div: Optional[str],
    exponent: Optional[int],
) -> Optional[str]:
    value = str(convert_float(raw_value, raw_div))
    if value is not None and exponent is not None:
        major, minor = value.split('.', 1)
        if exponent == 0 and minor == '0':
            return major
        missing = exponent - len(minor)
        if missing > 0:
            value += '0' * missing
    return value


def convert_timedelta(raw: Optional[str]) -> Optional[timedelta]:
    if raw is not None:
        return timedelta(seconds=int(raw, 0))
    return None


class ConnectionState(str, Enum):
    """Indicates the current state of the device."""

    def __str__(self) -> str:
        return super().__str__()

    INITIALIZING = 'Initializing'
    NETWORK_DISCOVERY = 'Network Discovery'
    JOINING = 'Joining'
    JOIN_FAIL = 'Join: Fail'
    JOIN_SUCCESS = 'Join: Success'
    AUTHENTICATING = 'Authenticating'
    AUTHENTICATING_SUCCESS = 'Authenticating: Success'
    AUTHENTICATING_FAIL = 'Authenticating: Fail'
    CONNECTED = 'Connected'
    DISCONNECTED = 'Disconnected'
    REJOINING = 'Rejoining'


class DataStatus(IntEnum):
    """Indicates the status of returned data."""

    SUCCESS = 0
    UNDEFINED = 1
    NOT_SUPPORTED = 2
    INVALID_END_TIME = 3
    TOO_MANY = 4
    NONE_AVAILABLE = 5


class IntervalChannel(str, Enum):
    """Profile interval data channels."""

    def __str__(self) -> str:
        return super().__str__()

    DELIVERED = 'Delivered'
    RECEIVED = 'Received'


class IntervalPeriod(IntEnum):
    """Sampling intervals."""

    DAILY = 0
    SIXTY_MIN = 1
    THIRTY_MIN = 2
    FIFTEEN_MIN = 3
    TEN_MIN = 4
    SEVEN_AND_A_HALF_MIN = 5
    FIVE_MIN = 6
    TWO_AND_A_HALF_MIN = 7


class MessageQueue(str, Enum):
    """Types of queues messages can occupy."""

    def __str__(self) -> str:
        return super().__str__()

    ACTIVE = 'Active'
    CANCEL_PENDING = 'Cancel Pending'


class MeterType(str, Enum):
    """Types of meters to connect to."""

    def __str__(self) -> str:
        return super().__str__()

    ELECTRIC = 'electric'
    GAS = 'gas'
    WATER = 'water'
    OTHER = 'other'


class ScheduledEvent(str, Enum):
    """Types of events which can be scheduled."""

    def __str__(self) -> str:
        return super().__str__()

    TIME = 'time'
    PRICE = 'price'
    DEMAND = 'demand'
    SUMMATION = 'summation'
    MESSAGE = 'message'


@dataclass
class ConnectionStatus:
    """Diagnostic information about the meter connection."""

    device_mac_id: Optional[bytes]
    coord_mac_id: Optional[bytes]
    status: Optional[ConnectionState]
    description: Optional[str]
    status_code: Optional[bytes]
    ext_pan_id: Optional[bytes]
    channel: Optional[int]
    short_addr: Optional[bytes]
    link_strength: Optional[int]


@dataclass
class CurrentPeriodUsage:
    """Total consumption for current accumulation period."""

    device_mac_id: Optional[bytes]
    meter_mac_id: Optional[bytes]
    time_stamp: Optional[datetime]
    current_usage: Optional[str]
    start_date: Optional[datetime]


@dataclass
class CurrentSummationDelivered:
    """Total consumption at the meter to date."""

    device_mac_id: Optional[bytes]
    meter_mac_id: Optional[bytes]
    time_stamp: Optional[datetime]
    summation_delivered: Optional[str]
    summation_received: Optional[str]


@dataclass
class DeviceInfo:
    """Information about the device."""

    device_mac_id: Optional[bytes]
    install_code: Optional[bytes]
    link_key: Optional[bytes]
    fw_version: Optional[str]
    hw_version: Optional[str]
    image_type: Optional[str]
    manufacturer: Optional[str]
    model_id: Optional[str]
    date_code: Optional[Tuple[date, str]]


@dataclass
class InstantaneousDemand:
    """Current consumption rate at the meter."""

    device_mac_id: Optional[bytes]
    meter_mac_id: Optional[bytes]
    time_stamp: Optional[datetime]
    demand: Optional[str]


@dataclass
class LastPeriodUsage:
    """Total consumption for the previous accumulation period."""

    device_mac_id: Optional[bytes]
    meter_mac_id: Optional[bytes]
    last_usage: Optional[str]
    start_date: Optional[datetime]
    end_date: Optional[datetime]


@dataclass
class MessageCluster:
    """Text messages from the meter."""

    device_mac_id: Optional[bytes]
    meter_mac_id: Optional[bytes]
    time_stamp: Optional[datetime]
    message_id: Optional[bytes]
    text: Optional[str]
    confirmation_required: Optional[bool]
    confirmed: Optional[bool]
    queue: Optional[MessageQueue]


@dataclass
class MeterInfo:
    """Information about a meter on the network."""

    device_mac_id: Optional[bytes]
    meter_mac_id: Optional[bytes]
    meter_type: Optional[MeterType]
    nick_name: Optional[str]
    account: Optional[str]
    auth: Optional[str]
    host: Optional[str]
    enabled: Optional[bool]


@dataclass
class MeterList:
    """List of meters the device is connected to."""

    device_mac_id: Optional[bytes]
    meter_mac_ids: Optional[List[bytes]]


@dataclass
class NetworkInfo(ConnectionStatus):
    """Information about the network the device is on."""

    device_mac_id: Optional[bytes]
    coord_mac_id: Optional[bytes]
    status: Optional[ConnectionState]
    description: Optional[str]
    status_code: Optional[bytes]
    ext_pan_id: Optional[bytes]
    channel: Optional[int]
    short_addr: Optional[bytes]
    link_strength: Optional[int]


@dataclass
class PriceCluster:
    """The current price in effect on the meter."""

    device_mac_id: Optional[bytes]
    meter_mac_id: Optional[bytes]
    time_stamp: Optional[datetime]
    price: Optional[str]
    currency: Optional[Currency]
    tier: Optional[int]
    tier_label: Optional[str]
    rate_label: Optional[str]
    # TODO(cottsay): Add 'Duration'


@dataclass
class ProfileData:
    """A series of interval data as recorded by the meter."""

    device_mac_id: Optional[bytes]
    meter_mac_id: Optional[bytes]
    end_time: Optional[datetime]
    status: Optional[DataStatus]
    profile_interval_period: Optional[IntervalPeriod]


@dataclass
class ScheduleInfo:
    """Information about periodic notifications sent by the device."""

    device_mac_id: Optional[bytes]
    meter_mac_id: Optional[bytes]
    event: Optional[ScheduledEvent]
    frequency: Optional[timedelta]
    enabled: Optional[bool]


@dataclass
class TimeCluster:
    """The current time reported on the meter."""

    device_mac_id: Optional[bytes]
    meter_mac_id: Optional[bytes]
    utc_time: Optional[datetime]
    local_time: Optional[datetime]



# Copyright 2022 Scott K Logan
# Licensed under the Apache License, Version 2.0

import asyncio
from datetime import date
from datetime import datetime
from datetime import timedelta
from datetime import timezone
import warnings

from aioraven.data import ConnectionState
from aioraven.data import CurrentPeriodUsage
from aioraven.data import CurrentSummationDelivered
from aioraven.data import DataStatus
from aioraven.data import DeviceInfo
from aioraven.data import InstantaneousDemand
from aioraven.data import IntervalChannel
from aioraven.data import IntervalPeriod
from aioraven.data import LastPeriodUsage
from aioraven.data import MessageCluster
from aioraven.data import MessageQueue
from aioraven.data import MeterInfo
from aioraven.data import MeterList
from aioraven.data import MeterType
from aioraven.data import NetworkInfo
from aioraven.data import PriceCluster
from aioraven.data import ProfileData
from aioraven.data import ScheduledEvent
from aioraven.data import ScheduleInfo
from aioraven.data import TimeCluster
from aioraven.device import RAVEnConnectionError
from aioraven.device import RAVEnWarning
from aioraven.device import UnknownRAVEnCommandWarning
from aioraven.streams import RAVEnNetworkDevice
from iso4217 import Currency
import pytest

from .mock_device import mock_device


@pytest.mark.asyncio
@pytest.mark.parametrize('meter', (bytes.fromhex('FEDCBA9876543210'), None))
async def test_close_current_period(meter):
    """Verify behavior of the ``close_current_period`` command."""
    responses = {
        b'<Command><Name>close_current_period</Name></Command>': None,
        b'<Command>'
        b'<Name>close_current_period</Name>'
        b'<MeterMacId>0xFEDCBA9876543210</MeterMacId>'
        b'</Command>': None,
    }

    async with mock_device(responses) as (host, port):
        async with RAVEnNetworkDevice(host, port) as dut:
            await dut.close_current_period(meter=meter)

    assert 1 == len(responses)


@pytest.mark.asyncio
@pytest.mark.parametrize('meter', (bytes.fromhex('FEDCBA9876543210'), None))
async def test_get_current_period_usage(meter):
    """Verify behavior of the ``get_current_period_usage`` command."""
    responses = {
        b'<Command><Name>get_current_period_usage</Name></Command>':
            b'<CurrentPeriodUsage>'
            b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
            b'    <MeterMacId>0xFEDCBA9876543210</MeterMacId>'
            b'    <TimeStamp>0x29bd58a7</TimeStamp>'
            b'    <CurrentUsage>0x00000010</CurrentUsage>'
            b'    <Multiplier>0x00000004</Multiplier>'
            b'    <Divisor>0x00000002</Divisor>'
            b'    <DigitsRight>0x02</DigitsRight>'
            b'    <DigitsLeft>0x04</DigitsLeft>'
            b'    <SuppressLeadingZero>N</SuppressLeadingZero>'
            b'    <StartDate>0x29AAE3A7</StartDate>'
            b'</CurrentPeriodUsage>',
    }
    responses[
        b'<Command>'
        b'<Name>get_current_period_usage</Name>'
        b'<MeterMacId>0xFEDCBA9876543210</MeterMacId>'
        b'</Command>'
    ] = responses[b'<Command><Name>get_current_period_usage</Name></Command>']

    async with mock_device(responses) as (host, port):
        async with RAVEnNetworkDevice(host, port) as dut:
            actual = await dut.get_current_period_usage(
                meter=meter,
            )

    assert actual == CurrentPeriodUsage(
        device_mac_id=bytes.fromhex('0123456789ABCDEF'),
        meter_mac_id=bytes.fromhex('FEDCBA9876543210'),
        time_stamp=datetime(
            2022, 3, 11, 0, 47, 35, tzinfo=timezone.utc),
        current_usage='0032.00',
        start_date=datetime(
            2022, 2, 25, 0, 47, 35, tzinfo=timezone.utc))


@pytest.mark.asyncio
@pytest.mark.parametrize('meter', (bytes.fromhex('FEDCBA9876543210'), None))
async def test_get_current_summation_delivered(meter):
    """Verify behavior of the ``get_current_summation_delivered`` command."""
    responses = {
        b'<Command><Name>get_current_summation_delivered</Name></Command>':
            b'<CurrentSummationDelivered>'
            b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
            b'    <MeterMacId>0xFEDCBA9876543210</MeterMacId>'
            b'    <TimeStamp>0x29bd58a7</TimeStamp>'
            b'    <SummationDelivered>0x00000010</SummationDelivered>'
            b'    <SummationReceived>0x00000008</SummationReceived>'
            b'    <Multiplier>0x00000004</Multiplier>'
            b'    <Divisor>0x00000002</Divisor>'
            b'    <DigitsRight>0x02</DigitsRight>'
            b'    <DigitsLeft>0x04</DigitsLeft>'
            b'    <SuppressLeadingZero>N</SuppressLeadingZero>'
            b'</CurrentSummationDelivered>',
    }
    responses[
        b'<Command>'
        b'<Name>get_current_summation_delivered</Name>'
        b'<MeterMacId>0xFEDCBA9876543210</MeterMacId>'
        b'</Command>'
    ] = responses[
        b'<Command><Name>get_current_summation_delivered</Name></Command>']

    async with mock_device(responses) as (host, port):
        async with RAVEnNetworkDevice(host, port) as dut:
            actual = await dut.get_current_summation_delivered(meter=meter)

    assert actual == CurrentSummationDelivered(
        device_mac_id=bytes.fromhex('0123456789ABCDEF'),
        meter_mac_id=bytes.fromhex('FEDCBA9876543210'),
        time_stamp=datetime(
            2022, 3, 11, 0, 47, 35, tzinfo=timezone.utc),
        summation_delivered='0032.00',
        summation_received='0016.00')


@pytest.mark.asyncio
async def test_get_current_summation_delivered_rounding():
    """
    Verify rounding behavior of the ``get_current_summation_delivered``
    command.
    """
    responses = {
        b'<Command><Name>get_current_summation_delivered</Name></Command>':
            b'<CurrentSummationDelivered>'
            b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
            b'    <MeterMacId>0xFEDCBA9876543210</MeterMacId>'
            b'    <TimeStamp>0x29bd58a7</TimeStamp>'
            b'    <SummationDelivered>0x00000C7B</SummationDelivered>'
            b'    <SummationReceived>0x00000644</SummationReceived>'
            b'    <Multiplier>0x00000002</Multiplier>'
            b'    <Divisor>0x000000C8</Divisor>'
            b'    <DigitsRight>0x01</DigitsRight>'
            b'    <DigitsLeft>0x03</DigitsLeft>'
            b'    <SuppressLeadingZero>N</SuppressLeadingZero>'
            b'</CurrentSummationDelivered>',
    }

    async with mock_device(responses) as (host, port):
        async with RAVEnNetworkDevice(host, port) as dut:
            actual = await dut.get_current_summation_delivered()

    assert actual == CurrentSummationDelivered(
        device_mac_id=bytes.fromhex('0123456789ABCDEF'),
        meter_mac_id=bytes.fromhex('FEDCBA9876543210'),
        time_stamp=datetime(
            2022, 3, 11, 0, 47, 35, tzinfo=timezone.utc),
        summation_delivered='032.0',
        summation_received='016.0')


@pytest.mark.asyncio
async def test_get_current_summation_delivered_no_received():
    """
    Verify behavior of the ``get_current_summation_delivered`` command without
    SummationReceived.
    """
    responses = {
        b'<Command><Name>get_current_summation_delivered</Name></Command>':
            b'<CurrentSummationDelivered>'
            b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
            b'    <MeterMacId>0xFEDCBA9876543210</MeterMacId>'
            b'    <TimeStamp>0x29bd58a7</TimeStamp>'
            b'    <SummationDelivered>0x00000010</SummationDelivered>'
            b'    <Multiplier>0x00000004</Multiplier>'
            b'    <Divisor>0x00000002</Divisor>'
            b'    <DigitsRight>0x02</DigitsRight>'
            b'    <DigitsLeft>0x04</DigitsLeft>'
            b'    <SuppressLeadingZero>N</SuppressLeadingZero>'
            b'</CurrentSummationDelivered>',
    }

    async with mock_device(responses) as (host, port):
        async with RAVEnNetworkDevice(host, port) as dut:
            actual = await dut.get_current_summation_delivered()

    assert actual == CurrentSummationDelivered(
        device_mac_id=bytes.fromhex('0123456789ABCDEF'),
        meter_mac_id=bytes.fromhex('FEDCBA9876543210'),
        time_stamp=datetime(
            2022, 3, 11, 0, 47, 35, tzinfo=timezone.utc),
        summation_delivered='0032.00',
        summation_received=None)


@pytest.mark.asyncio
@pytest.mark.parametrize('meter', (bytes.fromhex('FEDCBA9876543210'), None))
async def test_get_current_price(meter):
    """Verify behavior of the ``get_current_price`` command."""
    responses = {
        b'<Command><Name>get_current_price</Name></Command>':
            b'<PriceCluster>'
            b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
            b'    <MeterMacId>0xFEDCBA9876543210</MeterMacId>'
            b'    <TimeStamp>0x29bd58a7</TimeStamp>'
            b'    <Price>0xc8</Price>'
            b'    <Currency>0x348</Currency>'
            b'    <TrailingDigits>0x03</TrailingDigits>'
            b'    <Tier>0x08</Tier>'
            b'    <TierLabel>Set by User</TierLabel>'
            b'    <RateLabel>Set by User</RateLabel>'
            b'</PriceCluster>',
    }
    responses[
        b'<Command>'
        b'<Name>get_current_price</Name>'
        b'<MeterMacId>0xFEDCBA9876543210</MeterMacId>'
        b'</Command>'
    ] = responses[b'<Command><Name>get_current_price</Name></Command>']

    async with mock_device(responses) as (host, port):
        async with RAVEnNetworkDevice(host, port) as dut:
            actual = await dut.get_current_price(meter=meter)

    assert actual == PriceCluster(
        device_mac_id=bytes.fromhex('0123456789ABCDEF'),
        meter_mac_id=bytes.fromhex('FEDCBA9876543210'),
        time_stamp=datetime(
            2022, 3, 11, 0, 47, 35, tzinfo=timezone.utc),
        price='0.20',
        currency=Currency.usd,
        tier=8,
        tier_label='Set by User',
        rate_label='Set by User')


@pytest.mark.asyncio
async def test_get_current_price_no_currency():
    """
    Verify behavior of the ``get_current_price`` command without Currency.
    """
    responses = {
        b'<Command><Name>get_current_price</Name></Command>':
            b'<PriceCluster>'
            b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
            b'    <MeterMacId>0xFEDCBA9876543210</MeterMacId>'
            b'    <TimeStamp>0x29bd58a7</TimeStamp>'
            b'    <Price>0xc8</Price>'
            b'    <TrailingDigits>0x03</TrailingDigits>'
            b'    <Tier>0x08</Tier>'
            b'    <TierLabel>Set by User</TierLabel>'
            b'    <RateLabel>Set by User</RateLabel>'
            b'</PriceCluster>',
    }

    async with mock_device(responses) as (host, port):
        async with RAVEnNetworkDevice(host, port) as dut:
            actual = await dut.get_current_price()

    assert actual == PriceCluster(
        device_mac_id=bytes.fromhex('0123456789ABCDEF'),
        meter_mac_id=bytes.fromhex('FEDCBA9876543210'),
        time_stamp=datetime(
            2022, 3, 11, 0, 47, 35, tzinfo=timezone.utc),
        price='0.2',
        currency=None,
        tier=8,
        tier_label='Set by User',
        rate_label='Set by User')


@pytest.mark.asyncio
async def test_get_current_price_no_time_stamp():
    """
    Verify behavior of the ``get_current_price`` command without TimeStamp.
    """
    responses = {
        b'<Command><Name>get_current_price</Name></Command>':
            b'<PriceCluster>'
            b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
            b'    <MeterMacId>0xFEDCBA9876543210</MeterMacId>'
            b'    <Price>0xc8</Price>'
            b'    <Currency>0x348</Currency>'
            b'    <TrailingDigits>0x03</TrailingDigits>'
            b'    <Tier>0x08</Tier>'
            b'    <TierLabel>Set by User</TierLabel>'
            b'    <RateLabel>Set by User</RateLabel>'
            b'</PriceCluster>',
    }

    async with mock_device(responses) as (host, port):
        async with RAVEnNetworkDevice(host, port) as dut:
            actual = await dut.get_current_price()

    assert actual == PriceCluster(
        device_mac_id=bytes.fromhex('0123456789ABCDEF'),
        meter_mac_id=bytes.fromhex('FEDCBA9876543210'),
        time_stamp=None,
        price='0.20',
        currency=Currency.usd,
        tier=8,
        tier_label='Set by User',
        rate_label='Set by User')


@pytest.mark.asyncio
async def test_get_device_info():
    """Verify behavior of the ``get_device_info`` command."""
    responses = {
        b'<Command><Name>get_device_info</Name></Command>':
            b'<DeviceInfo>'
            b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
            b'    <InstallCode>0xABCDEF0123456789</InstallCode>'
            b'    <LinkKey>0xABCDEF0123456789ABCDEF0123456789</LinkKey>'
            b'    <FWVersion>1.21g</FWVersion>'
            b'    <HWVersion>5.55 rev 2</HWVersion>'
            b'    <ImageType>Mocked</ImageType>'
            b'    <Manufacturer>aioraven</Manufacturer>'
            b'    <ModelId>Python</ModelId>'
            b'    <DateCode>20220101a0000042</DateCode>'
            b'</DeviceInfo>',
    }

    async with mock_device(responses) as (host, port):
        async with RAVEnNetworkDevice(host, port) as dut:
            actual = await dut.get_device_info()

    assert actual == DeviceInfo(
        device_mac_id=bytes.fromhex('0123456789ABCDEF'),
        install_code=bytes.fromhex('ABCDEF0123456789'),
        link_key=bytes.fromhex('ABCDEF0123456789ABCDEF0123456789'),
        fw_version='1.21g',
        hw_version='5.55 rev 2',
        image_type='Mocked',
        manufacturer='aioraven',
        model_id='Python',
        date_code=(date(2022, 1, 1), 'a0000042'))


@pytest.mark.asyncio
async def test_get_instantaneous_demand():
    """Verify behavior of the ``get_instantaneous_demand`` command."""
    responses = {
        b'<Command><Name>get_instantaneous_demand</Name></Command>':
            b'<InstantaneousDemand>'
            b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
            b'    <MeterMacId>0xFEDCBA9876543210</MeterMacId>'
            b'    <TimeStamp>0x29bd58a7</TimeStamp>'
            b'    <Demand>0x00000010</Demand>'
            b'    <Multiplier>0x00000004</Multiplier>'
            b'    <Divisor>0x00000002</Divisor>'
            b'    <DigitsRight>0x02</DigitsRight>'
            b'    <DigitsLeft>0x04</DigitsLeft>'
            b'    <SuppressLeadingZero>Y</SuppressLeadingZero>'
            b'</InstantaneousDemand>',
    }

    async with mock_device(responses) as (host, port):
        async with RAVEnNetworkDevice(host, port) as dut:
            actual = await dut.get_instantaneous_demand()

    assert actual == InstantaneousDemand(
        device_mac_id=bytes.fromhex('0123456789ABCDEF'),
        meter_mac_id=bytes.fromhex('FEDCBA9876543210'),
        time_stamp=datetime(
            2022, 3, 11, 0, 47, 35, tzinfo=timezone.utc),
        demand='32.00')


@pytest.mark.asyncio
async def test_get_instantaneous_demand_negative():
    """
    Verify behavior of the ``get_instantaneous_demand`` command with a
    negative value.
    """
    responses = {
        b'<Command><Name>get_instantaneous_demand</Name></Command>':
            b'<InstantaneousDemand>'
            b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
            b'    <MeterMacId>0xFEDCBA9876543210</MeterMacId>'
            b'    <TimeStamp>0x29bd58a7</TimeStamp>'
            b'    <Demand>0xFFFFFFF0</Demand>'
            b'    <Multiplier>0x00000004</Multiplier>'
            b'    <Divisor>0x00000002</Divisor>'
            b'    <DigitsRight>0x02</DigitsRight>'
            b'    <DigitsLeft>0x04</DigitsLeft>'
            b'    <SuppressLeadingZero>Y</SuppressLeadingZero>'
            b'</InstantaneousDemand>',
    }

    async with mock_device(responses) as (host, port):
        async with RAVEnNetworkDevice(host, port) as dut:
            actual = await dut.get_instantaneous_demand()

    assert actual == InstantaneousDemand(
        device_mac_id=bytes.fromhex('0123456789ABCDEF'),
        meter_mac_id=bytes.fromhex('FEDCBA9876543210'),
        time_stamp=datetime(
            2022, 3, 11, 0, 47, 35, tzinfo=timezone.utc),
        demand='-32.00')


@pytest.mark.asyncio
async def test_get_instantaneous_demand_negative_no_slz():
    """
    Verify behavior of the ``get_instantaneous_demand`` command with a
    negative value and without SuppressLeadingZero.
    """
    responses = {
        b'<Command><Name>get_instantaneous_demand</Name></Command>':
            b'<InstantaneousDemand>'
            b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
            b'    <MeterMacId>0xFEDCBA9876543210</MeterMacId>'
            b'    <TimeStamp>0x29bd58a7</TimeStamp>'
            b'    <Demand>0xFFFFFFF0</Demand>'
            b'    <Multiplier>0x00000004</Multiplier>'
            b'    <Divisor>0x00000002</Divisor>'
            b'    <DigitsRight>0x02</DigitsRight>'
            b'    <DigitsLeft>0x04</DigitsLeft>'
            b'</InstantaneousDemand>',
    }

    async with mock_device(responses) as (host, port):
        async with RAVEnNetworkDevice(host, port) as dut:
            actual = await dut.get_instantaneous_demand()

    assert actual == InstantaneousDemand(
        device_mac_id=bytes.fromhex('0123456789ABCDEF'),
        meter_mac_id=bytes.fromhex('FEDCBA9876543210'),
        time_stamp=datetime(
            2022, 3, 11, 0, 47, 35, tzinfo=timezone.utc),
        demand='-0032.00')


@pytest.mark.asyncio
async def test_get_instantaneous_demand_no_slz():
    """
    Verify behavior of the ``get_instantaneous_demand`` command without
    SuppressLeadingZero.
    """
    responses = {
        b'<Command><Name>get_instantaneous_demand</Name></Command>':
            b'<InstantaneousDemand>'
            b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
            b'    <MeterMacId>0xFEDCBA9876543210</MeterMacId>'
            b'    <TimeStamp>0x29bd58a7</TimeStamp>'
            b'    <Demand>0x00000010</Demand>'
            b'    <Multiplier>0x00000004</Multiplier>'
            b'    <Divisor>0x00000002</Divisor>'
            b'    <DigitsRight>0x02</DigitsRight>'
            b'    <DigitsLeft>0x04</DigitsLeft>'
            b'</InstantaneousDemand>',
    }

    async with mock_device(responses) as (host, port):
        async with RAVEnNetworkDevice(host, port) as dut:
            actual = await dut.get_instantaneous_demand()

    assert actual == InstantaneousDemand(
        device_mac_id=bytes.fromhex('0123456789ABCDEF'),
        meter_mac_id=bytes.fromhex('FEDCBA9876543210'),
        time_stamp=datetime(
            2022, 3, 11, 0, 47, 35, tzinfo=timezone.utc),
        demand='0032.00')


@pytest.mark.asyncio
async def test_get_instantaneous_demand_no_digits_right():
    """
    Verify behavior of the ``get_instantaneous_demand`` command without
    DigitsRight.
    """
    responses = {
        b'<Command><Name>get_instantaneous_demand</Name></Command>':
            b'<InstantaneousDemand>'
            b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
            b'    <MeterMacId>0xFEDCBA9876543210</MeterMacId>'
            b'    <TimeStamp>0x29bd58a7</TimeStamp>'
            b'    <Demand>0x00000010</Demand>'
            b'    <Multiplier>0x00000004</Multiplier>'
            b'    <Divisor>0x00000002</Divisor>'
            b'    <DigitsLeft>0x04</DigitsLeft>'
            b'    <SuppressLeadingZero>N</SuppressLeadingZero>'
            b'</InstantaneousDemand>',
    }

    async with mock_device(responses) as (host, port):
        async with RAVEnNetworkDevice(host, port) as dut:
            actual = await dut.get_instantaneous_demand()

    assert actual == InstantaneousDemand(
        device_mac_id=bytes.fromhex('0123456789ABCDEF'),
        meter_mac_id=bytes.fromhex('FEDCBA9876543210'),
        time_stamp=datetime(
            2022, 3, 11, 0, 47, 35, tzinfo=timezone.utc),
        demand='0032.0')


@pytest.mark.asyncio
async def test_get_instantaneous_demand_no_digits_left():
    """
    Verify behavior of the ``get_instantaneous_demand`` command without
    DigitsLeft.
    """
    responses = {
        b'<Command><Name>get_instantaneous_demand</Name></Command>':
            b'<InstantaneousDemand>'
            b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
            b'    <MeterMacId>0xFEDCBA9876543210</MeterMacId>'
            b'    <TimeStamp>0x29bd58a7</TimeStamp>'
            b'    <Demand>0x00000010</Demand>'
            b'    <Multiplier>0x00000004</Multiplier>'
            b'    <Divisor>0x00000002</Divisor>'
            b'    <DigitsRight>0x02</DigitsRight>'
            b'    <SuppressLeadingZero>N</SuppressLeadingZero>'
            b'</InstantaneousDemand>',
    }

    async with mock_device(responses) as (host, port):
        async with RAVEnNetworkDevice(host, port) as dut:
            actual = await dut.get_instantaneous_demand()

    assert actual == InstantaneousDemand(
        device_mac_id=bytes.fromhex('0123456789ABCDEF'),
        meter_mac_id=bytes.fromhex('FEDCBA9876543210'),
        time_stamp=datetime(
            2022, 3, 11, 0, 47, 35, tzinfo=timezone.utc),
        demand='32.00')


@pytest.mark.asyncio
async def test_get_last_period_usage():
    """Verify behavior of the ``get_last_period_usage`` command."""
    responses = {
        b'<Command><Name>get_last_period_usage</Name></Command>':
            b'<LastPeriodUsage>'
            b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
            b'    <MeterMacId>0xFEDCBA9876543210</MeterMacId>'
            b'    <LastUsage>0x00000010</LastUsage>'
            b'    <Multiplier>0x00000004</Multiplier>'
            b'    <Divisor>0x00000002</Divisor>'
            b'    <DigitsRight>0x02</DigitsRight>'
            b'    <DigitsLeft>0x04</DigitsLeft>'
            b'    <SuppressLeadingZero>N</SuppressLeadingZero>'
            b'    <StartDate>0x29bd58a7</StartDate>'
            b'    <EndDate>0x29E63727</EndDate>'
            b'</LastPeriodUsage>',
    }

    async with mock_device(responses) as (host, port):
        async with RAVEnNetworkDevice(host, port) as dut:
            actual = await dut.get_last_period_usage()

    assert actual == LastPeriodUsage(
        device_mac_id=bytes.fromhex('0123456789ABCDEF'),
        meter_mac_id=bytes.fromhex('FEDCBA9876543210'),
        last_usage='0032.00',
        start_date=datetime(
            2022, 3, 11, 0, 47, 35, tzinfo=timezone.utc),
        end_date=datetime(
            2022, 4, 11, 0, 47, 35, tzinfo=timezone.utc))


@pytest.mark.asyncio
async def test_get_message():
    """Verify behavior of the ``get_message`` command."""
    responses = {
        b'<Command><Name>get_message</Name></Command>':
            b'<MessageCluster>'
            b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
            b'    <MeterMacId>0xFEDCBA9876543210</MeterMacId>'
            b'    <TimeStamp>0x29bd58a7</TimeStamp>'
            b'    <Id>0x02468ACE</Id>'
            b'    <Text>Hello, World!</Text>'
            b'    <ConfirmationRequired>N</ConfirmationRequired>'
            b'    <Confirmed>Y</Confirmed>'
            b'    <Queue>Active</Queue>'
            b'</MessageCluster>',
    }

    async with mock_device(responses) as (host, port):
        async with RAVEnNetworkDevice(host, port) as dut:
            actual = await dut.get_message()

    assert actual == MessageCluster(
        device_mac_id=bytes.fromhex('0123456789ABCDEF'),
        meter_mac_id=bytes.fromhex('FEDCBA9876543210'),
        time_stamp=datetime(
            2022, 3, 11, 0, 47, 35, tzinfo=timezone.utc),
        message_id=bytes.fromhex('02468ACE'),
        text='Hello, World!',
        confirmation_required=False,
        confirmed=True,
        queue=MessageQueue.ACTIVE)


@pytest.mark.asyncio
async def test_get_meter_info():
    """Verify behavior of the ``get_meter_info`` command."""
    responses = {
        b'<Command><Name>get_meter_info</Name></Command>':
            b'<MeterInfo>'
            b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
            b'    <MeterMacId>0xFEDCBA9876543210</MeterMacId>'
            b'    <MeterType>electric</MeterType>'
            b'    <NickName>House</NickName>'
            b'    <Account>8675309</Account>'
            b'    <Auth>p@ssw0rd</Auth>'
            b'    <Host>Example, Inc.</Host>'
            b'    <Enabled>Y</Enabled>'
            b'</MeterInfo>',
    }

    async with mock_device(responses) as (host, port):
        async with RAVEnNetworkDevice(host, port) as dut:
            actual = await dut.get_meter_info()

    assert actual == MeterInfo(
        device_mac_id=bytes.fromhex('0123456789ABCDEF'),
        meter_mac_id=bytes.fromhex('FEDCBA9876543210'),
        meter_type=MeterType.ELECTRIC,
        nick_name='House',
        account='8675309',
        auth='p@ssw0rd',
        host='Example, Inc.',
        enabled=True)


@pytest.mark.asyncio
async def test_get_network_info():
    """Verify behavior of the ``get_network_info`` command."""
    responses = {
        b'<Command><Name>get_network_info</Name></Command>':
            b'<NetworkInfo>'
            b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
            b'    <CoordMacId>0xFEDCBA9876543210</CoordMacId>'
            b'    <Status>Connected</Status>'
            b'    <Description>Network is operational</Description>'
            b'    <StatusCode>0x42</StatusCode>'
            b'    <ExtPanId>0x9876543210ABCDEF</ExtPanId>'
            b'    <Channel>24</Channel>'
            b'    <ShortAddr>0x5678</ShortAddr>'
            b'    <LinkStrength>0x24</LinkStrength>'
            b'</NetworkInfo>',
    }

    async with mock_device(responses) as (host, port):
        async with RAVEnNetworkDevice(host, port) as dut:
            actual = await dut.get_network_info()

    assert actual == NetworkInfo(
        device_mac_id=bytes.fromhex('0123456789ABCDEF'),
        coord_mac_id=bytes.fromhex('FEDCBA9876543210'),
        status=ConnectionState.CONNECTED,
        description='Network is operational',
        status_code=bytes.fromhex('42'),
        ext_pan_id=bytes.fromhex('9876543210ABCDEF'),
        channel=24,
        short_addr=bytes.fromhex('5678'),
        link_strength=36)


@pytest.mark.asyncio
async def test_get_meter_list_zero():
    """Verify behavior of the ``get_meter_list`` command."""
    responses = {
        b'<Command><Name>get_meter_list</Name></Command>':
            b'<MeterList>'
            b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
            b'</MeterList>',
    }

    async with mock_device(responses) as (host, port):
        async with RAVEnNetworkDevice(host, port) as dut:
            actual = await dut.get_meter_list()

    assert actual == MeterList(
        device_mac_id=bytes.fromhex('0123456789ABCDEF'),
        meter_mac_ids=[])


@pytest.mark.asyncio
async def test_get_meter_list_one():
    """Verify behavior of the ``get_meter_list`` command."""
    responses = {
        b'<Command><Name>get_meter_list</Name></Command>':
            b'<MeterList>'
            b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
            b'    <MeterMacId>0xFEDCBA9876543210</MeterMacId>'
            b'</MeterList>',
    }

    async with mock_device(responses) as (host, port):
        async with RAVEnNetworkDevice(host, port) as dut:
            actual = await dut.get_meter_list()

    assert actual == MeterList(
        device_mac_id=bytes.fromhex('0123456789ABCDEF'),
        meter_mac_ids=[bytes.fromhex('FEDCBA9876543210')])


@pytest.mark.asyncio
async def test_get_meter_list_two():
    """Verify behavior of the ``get_meter_list`` command."""
    responses = {
        b'<Command><Name>get_meter_list</Name></Command>':
            b'<MeterList>'
            b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
            b'    <MeterMacId>0xFEDCBA9876543210</MeterMacId>'
            b'    <MeterMacId>0xfedcba0123456789</MeterMacId>'
            b'</MeterList>',
    }

    async with mock_device(responses) as (host, port):
        async with RAVEnNetworkDevice(host, port) as dut:
            actual = await dut.get_meter_list()

    assert actual == MeterList(
        device_mac_id=bytes.fromhex('0123456789ABCDEF'),
        meter_mac_ids=[
            bytes.fromhex('FEDCBA9876543210'),
            bytes.fromhex('FEDCBA0123456789')])


@pytest.mark.asyncio
async def test_get_profile_data():
    """Verify behavior of the ``get_profile_data`` command."""
    responses = {
        b'<Command>'
        b'<Name>get_profile_data</Name>'
        b'<NumberOfPeriods>0x02</NumberOfPeriods>'
        b'<EndTime>0x29bd58a7</EndTime>'
        b'<IntervalChannel>Delivered</IntervalChannel>'
        b'</Command>':
            b'<ProfileData>'
            b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
            b'    <MeterMacId>0xFEDCBA9876543210</MeterMacId>'
            b'    <EndTime>0x29bd58a7</EndTime>'
            b'    <Status>0x00</Status>'
            b'    <ProfileIntervalPeriod>2</ProfileIntervalPeriod>'
            b'</ProfileData>',
    }

    async with mock_device(responses) as (host, port):
        async with RAVEnNetworkDevice(host, port) as dut:
            actual = await dut.get_profile_data(
                2, '0x29bd58a7',
                IntervalChannel.DELIVERED)

    assert actual == ProfileData(
        device_mac_id=bytes.fromhex('0123456789ABCDEF'),
        meter_mac_id=bytes.fromhex('FEDCBA9876543210'),
        end_time=datetime(
            2022, 3, 11, 0, 47, 35, tzinfo=timezone.utc),
        status=DataStatus.SUCCESS,
        profile_interval_period=IntervalPeriod.THIRTY_MIN)


@pytest.mark.asyncio
async def test_get_schedule():
    """Verify behavior of the ``get_schedule`` command."""
    responses = {
        b'<Command><Name>get_schedule</Name></Command>':
            b'<ScheduleInfo>'
            b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
            b'    <MeterMacId>0xFEDCBA9876543210</MeterMacId>'
            b'    <Event>summation</Event>'
            b'    <Frequency>0x13579bdf</Frequency>'
            b'    <Enabled>Y</Enabled>'
            b'</ScheduleInfo>',
    }

    async with mock_device(responses) as (host, port):
        async with RAVEnNetworkDevice(host, port) as dut:
            actual = await dut.get_schedule()

    assert actual == ScheduleInfo(
        device_mac_id=bytes.fromhex('0123456789ABCDEF'),
        meter_mac_id=bytes.fromhex('FEDCBA9876543210'),
        event=ScheduledEvent.SUMMATION,
        frequency=timedelta(seconds=0x13579bdf),
        enabled=True)


@pytest.mark.asyncio
async def test_get_time():
    """Verify behavior of the ``get_time`` command."""
    responses = {
        b'<Command><Name>get_time</Name></Command>':
            b'<TimeCluster>'
            b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
            b'    <MeterMacId>0xFEDCBA9876543210</MeterMacId>'
            b'    <UTCTime>0x29bd58a7</UTCTime>'
            b'    <LocalTime>0x29bce827</LocalTime>'
            b'</TimeCluster>',
    }

    async with mock_device(responses) as (host, port):
        async with RAVEnNetworkDevice(host, port) as dut:
            actual = await dut.get_time()

    assert actual == TimeCluster(
        device_mac_id=bytes.fromhex('0123456789ABCDEF'),
        meter_mac_id=bytes.fromhex('FEDCBA9876543210'),
        utc_time=datetime(
            2022, 3, 11, 0, 47, 35, tzinfo=timezone.utc),
        local_time=datetime(2022, 3, 10, 16, 47, 35))


@pytest.mark.asyncio
async def test_device_synchronize_clean():
    """Verify behavior of the ``synchronize`` helper."""
    responses = {
        b'<Command><Name>get_meter_list</Name></Command>':
            b'<MeterList>'
            b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
            b'</MeterList>',
    }

    async with mock_device(responses) as (host, port):
        async with RAVEnNetworkDevice(host, port) as dut:
            await dut.synchronize()


@pytest.mark.asyncio
async def test_device_synchronize_pre_leftovers():
    """Verify behavior of the ``synchronize`` helper."""
    responses = {
        b'<Command><Name>get_meter_list</Name></Command>':
            b'<MeterList>'
            b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
            b'</MeterList>',
    }

    async with mock_device(responses, b'</Leftovers>') as (host, port):
        async with RAVEnNetworkDevice(host, port) as dut:
            await dut.synchronize()


@pytest.mark.asyncio
async def test_device_synchronize_post_leftovers():
    """Verify behavior of the ``synchronize`` helper."""
    responses = {
        b'<Command><Name>get_meter_list</Name></Command>':
            [
                b'</Leftovers>\n'
                b'<MeterList>'
                b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
                b'</MeterList>',
                b'<MeterList>'
                b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
                b'</MeterList>',
            ],
    }

    async with mock_device(responses) as (host, port):
        async with RAVEnNetworkDevice(host, port) as dut:
            await dut.synchronize()


@pytest.mark.asyncio
async def test_device_synchronize_persistent_leftovers():
    """Verify behavior of the ``synchronize`` helper."""
    responses = {
        b'<Command><Name>get_meter_list</Name></Command>':
            [
                b'</Leftovers>\n'
                b'<MeterList>'
                b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
                b'</MeterList>',
            ] * 3,
    }

    async with mock_device(responses) as (host, port):
        async with RAVEnNetworkDevice(host, port) as dut:
            with pytest.raises(RAVEnConnectionError):
                await dut.synchronize()


@pytest.mark.asyncio
async def test_device_synchronize_timeouts():
    """Verify behavior of the 'synchronize' helper timeouts."""
    responses = {
        b'<Command><Name>get_meter_list</Name></Command>':
            [b''] * 4 + [
                b'<MeterList>'
                b'    <DeviceMacId>0x0123456789ABCDEF</DeviceMacId>'
                b'</MeterList>',
            ],
    }

    async with mock_device(responses) as (host, port):
        async with RAVEnNetworkDevice(host, port) as dut:
            with pytest.raises(asyncio.TimeoutError):
                await dut.synchronize(retries=3, timeout=0.05)
            await dut.synchronize()


@pytest.mark.asyncio
async def test_device_warning_generic():
    """Verify behavior of the ``get_meter_list`` command."""
    responses = {
        b'<Command><Name>get_meter_list</Name></Command>':
            b'<Warning>'
            b'    <Text>Something unexpected happened</Text>'
            b'</Warning>',
    }

    with warnings.catch_warnings(record=True) as w:
        async with mock_device(responses) as (host, port):
            async with RAVEnNetworkDevice(host, port) as dut:
                with pytest.raises(asyncio.TimeoutError):
                    await asyncio.wait_for(dut.get_meter_list(), timeout=0.05)

        assert len(w) == 1
        assert issubclass(w[-1].category, RAVEnWarning)
        assert 'Something unexpected happened' in str(w[-1].message)


@pytest.mark.asyncio
async def test_device_warning_generic_error():
    """Verify behavior of generic device warnings."""
    responses = {
        b'<Command><Name>get_meter_list</Name></Command>':
            b'<Warning>'
            b'    <Text>Something unexpected happened</Text>'
            b'</Warning>',
    }

    with warnings.catch_warnings():
        warnings.simplefilter('error', RAVEnWarning)
        async with mock_device(responses) as (host, port):
            async with RAVEnNetworkDevice(host, port) as dut:
                with pytest.raises(RAVEnWarning):
                    await dut.get_meter_list()


@pytest.mark.asyncio
async def test_device_warning_unknown_command():
    """Verify behavior of the ``Unknown command`` device warning."""
    responses = {
        b'<Command><Name>get_meter_list</Name></Command>':
            b'<Warning>'
            b'    <Text>Unknown command</Text>'
            b'</Warning>',
    }

    with warnings.catch_warnings():
        warnings.simplefilter('error', RAVEnWarning)
        async with mock_device(responses) as (host, port):
            async with RAVEnNetworkDevice(host, port) as dut:
                with pytest.raises(UnknownRAVEnCommandWarning):
                    await dut.get_meter_list()
